# -*- coding: utf-8 -*-
"""
    crimson.fusioncatcher
    ~~~~~~~~~~~~~~~~~~~~~

    FusionCatcher output parsing.

"""
# (c) 2015-2018 Wibowo Arindrarto <bow@bow.web.id>
import click

from .utils import get_handle


__all__ = ["parse"]

# Expected column names
_COLS = [
    "Gene_1_symbol(5end_fusion_partner)",
    "Gene_2_symbol(3end_fusion_partner)",
    "Fusion_description", "Counts_of_common_mapping_reads",
    "Spanning_pairs", "Spanning_unique_reads", "Longest_anchor_found",
    "Fusion_finding_method",
    "Fusion_point_for_gene_1(5end_fusion_partner)",
    "Fusion_point_for_gene_2(3end_fusion_partner)",
    "Gene_1_id(5end_fusion_partner)", "Gene_2_id(3end_fusion_partner)",
    "Exon_1_id(5end_fusion_partner)", "Exon_2_id(3end_fusion_partner)",
    "Fusion_sequence", "Predicted_effect", "Predicted_fused_transcripts",
    "Predicted_fused_proteins",
]

# Delimiter strings
_DELIM = {
    # Fusion description
    "desc": ",",
    # Chromosome-coordinate-strand
    "loc": ":",
    # Others
    "gen": ";",
}


def parse_lr_entry(lr_gene, lr_brkpoint):
    """Parses the gene and breakpoint entry.

    :param lr_gene: Column value for 'LeftGene' or 'RightGene'.
    :type lr_gene: str
    :param lr_brkpoint: Column value for 'LeftBreakpoint' or 'RightBreakpoint'.
    :type lr_brkpoint: str
    :rtype: dict

    """
    lrgname, lrgid = lr_gene.split(_DELIM["gids"])
    lrchrom, lrpos, lrstrand = lr_brkpoint.split(_DELIM["loc"])

    return {
        "geneName": lrgname,
        "geneID": lrgid,
        "chromosome": lrchrom,
        "position": int(lrpos),
        "strand": lrstrand,
    }


def split_filter(string, delim):
    if not string:
        return []
    return string.split(delim)


_COLS = [
    "Gene_1_symbol(5end_fusion_partner)",
    "Gene_2_symbol(3end_fusion_partner)",
    "Fusion_description", "Counts_of_common_mapping_reads",
    "Spanning_pairs", "Spanning_unique_reads", "Longest_anchor_found",
    "Fusion_finding_method",
    "Fusion_point_for_gene_1(5end_fusion_partner)",
    "Fusion_point_for_gene_2(3end_fusion_partner)",
    "Gene_1_id(5end_fusion_partner)", "Gene_2_id(3end_fusion_partner)",
    "Exon_1_id(5end_fusion_partner)", "Exon_2_id(3end_fusion_partner)",
    "Fusion_sequence", "Predicted_effect", "Predicted_fused_transcripts",
    "Predicted_fused_proteins",
]


def parse_raw_line(raw_line, colnames=_COLS):
    """Parses a single line into a dictionary.

    :param raw_line: FusionCatcher result line.
    :type raw_line: str
    :param colnames: Column names present in the file.
    :type colnames: list of str
    :rtype: dict

    """
    values = raw_line.split("\t")
    if len(values) != len(colnames):
        msg = "Line values {0} does not match column names {1}."
        raise click.BadParameter(msg.format(values, colnames))

    d = {k: v.strip() for k, v in zip(colnames, values)}

    f5 = d["Fusion_point_for_gene_1(5end_fusion_partner)"].split(_DELIM["loc"])
    f3 = d["Fusion_point_for_gene_2(3end_fusion_partner)"].split(_DELIM["loc"])

    res = {
        "5end": {
            "geneSymbol": d["Gene_1_symbol(5end_fusion_partner)"],
            "geneID": d["Gene_1_id(5end_fusion_partner)"],
            "exonID": d["Exon_1_id(5end_fusion_partner)"] or None,
            "chromosome": f5[0],
            "position": int(f5[1]),
            "strand": f5[2],
        },
        "3end": {
            "geneSymbol": d["Gene_2_symbol(3end_fusion_partner)"],
            "geneID": d["Gene_2_id(3end_fusion_partner)"],
            "exonID": d["Exon_2_id(3end_fusion_partner)"] or None,
            "chromosome": f3[0],
            "position": int(f3[1]),
            "strand": f3[2],
        },
        "fusionDescription":
            split_filter(d["Fusion_description"], _DELIM["desc"]),
        "nCommonMappingReads": int(d["Counts_of_common_mapping_reads"]),
        "nSpanningPairs": int(d["Spanning_pairs"]),
        "nSpanningUniqueReads": int(d["Spanning_unique_reads"]),
        "longestAnchorLength": int(d["Longest_anchor_found"]),
        "fusionFindingMethod":
            split_filter(d["Fusion_finding_method"], _DELIM["gen"]),
        "fusionSequence": d["Fusion_sequence"],
        "predictedEffect": d["Predicted_effect"],
        "predictedFusedTranscripts":
            split_filter(d["Predicted_fused_transcripts"], _DELIM["gen"]),
        "predictedFusedProteins":
            split_filter(d["Predicted_fused_proteins"], _DELIM["gen"]),
    }
    return res


def parse(in_data):
    """Parses the abridged output of a FusionCatcher run.

    :param in_data: Input FusionCatcher contents.
    :type in_data: str or file handle
    :returns: Parsed values.
    :rtype: dict

    """
    payload = []
    with get_handle(in_data) as src:
        first_line = src.readline().strip()
        # Parse column names
        colnames = first_line.split("\t")
        if colnames != _COLS:
            msg = "Unexpected column names: {0}."
            raise click.BadParameter(msg.format(colnames))
        for line in (x for x in src):
            parsed = parse_raw_line(line)
            payload.append(parsed)

    return payload



# -*- coding: utf-8 -*-
"""
    star-fusion subcommand tests
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
# (c) 2015-2018 Wibowo Arindrarto <bow@bow.web.id>
import json
import pytest
from click.testing import CliRunner

from crimson.cli import main

from .utils import get_test_path, getattr_nested


@pytest.fixture(scope="module")
def star_fusion_fail():
    runner = CliRunner()
    in_file = get_test_path("star_fusion_nope.txt")
    result = runner.invoke(main, ["star-fusion", in_file])
    return result


@pytest.fixture(scope="module")
def star_fusion_v060_01():
    runner = CliRunner()
    in_file = get_test_path("star_fusion_v060_01.txt")
    result = runner.invoke(main, ["star-fusion", in_file])
    result.json = json.loads(result.output)
    return result


@pytest.fixture(scope="module")
def star_fusion_v060_02():
    runner = CliRunner()
    in_file = get_test_path("star_fusion_v060_02.txt")
    result = runner.invoke(main, ["star-fusion", in_file])
    result.json = json.loads(result.output)
    return result


def test_star_fusion_fail_exit_code(star_fusion_fail):
    assert star_fusion_fail.exit_code != 0


def test_star_fusion_fail_output(star_fusion_fail):
    err_msg = "Unexpected header line:"
    assert err_msg in star_fusion_fail.output


@pytest.mark.parametrize("attrs, exp", [
    ([0, "fusionName"], "RUNX1--RUNX1T1"),
    ([0, "nJunctionReads"], 30),
    ([0, "nSpanningFrags"], 8),
    ([0, "spliceType"], "ONLY_REF_SPLICE"),
    ([0, "left", "geneName"], "RUNX1"),
    ([0, "left", "geneID"], "ENSG00000159216.18"),
    ([0, "left", "chromosome"], "chr21"),
    ([0, "left", "position"], 34859474),
    ([0, "left", "strand"], "-"),
    ([0, "right", "geneName"], "RUNX1T1"),
    ([0, "right", "geneID"], "ENSG00000079102.16"),
    ([0, "right", "chromosome"], "chr8"),
    ([0, "right", "position"], 92017363),
    ([0, "right", "strand"], "-"),
    ([-1, "fusionName"], "ITM2C--PTMA"),
    ([-1, "nJunctionReads"], 1),
    ([-1, "nSpanningFrags"], 1),
    ([-1, "spliceType"], "ONLY_REF_SPLICE"),
    ([-1, "left", "geneName"], "ITM2C"),
    ([-1, "left", "geneID"], "ENSG00000135916.15"),
    ([-1, "left", "chromosome"], "chr2"),
    ([-1, "left", "position"], 230865145),
    ([-1, "left", "strand"], "+"),
    ([-1, "right", "geneName"], "PTMA"),
    ([-1, "right", "geneID"], "ENSG00000187514.14"),
    ([-1, "right", "chromosome"], "chr2"),
    ([-1, "right", "position"], 231711348),
    ([-1, "right", "strand"], "+"),
])
def test_star_fusion_v060_01(star_fusion_v060_01, attrs, exp):
    assert getattr_nested(star_fusion_v060_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in attrs])


@pytest.mark.parametrize("attrs, exp", [
    ([0, "fusionName"], "KANSL1--ARL17B"),
    ([0, "nJunctionReads"], 23),
    ([0, "nSpanningFrags"], 4),
    ([0, "spliceType"], "ONLY_REF_SPLICE"),
    ([0, "left", "geneName"], "KANSL1"),
    ([0, "left", "geneID"], "ENSG00000120071.12"),
    ([0, "left", "chromosome"], "chr17"),
    ([0, "left", "position"], 46094560),
    ([0, "left", "strand"], "-"),
    ([0, "right", "geneName"], "ARL17B"),
    ([0, "right", "geneID"], "ENSG00000228696.8"),
    ([0, "right", "chromosome"], "chr17"),
    ([0, "right", "position"], 46352930),
    ([0, "right", "strand"], "-"),
    ([0, "reads", "junctionReads", 0],
     "HISEQ:113:C6ALHANXX:5:1308:16235:84862"),
    ([0, "reads", "junctionReads", -1],
     "HISEQ:114:C6DWCANXX:4:2102:20538:80656"),
    ([0, "reads", "spanningFrags", 0],
     "HISEQ:114:C6DWCANXX:4:1102:11126:100433"),
    ([0, "reads", "spanningFrags", -1],
     "HISEQ:114:C6DWCANXX:4:1106:20997:30657"),
    ([-1, "fusionName"], "NCOR2--UBC"),
    ([-1, "nJunctionReads"], 1),
    ([-1, "nSpanningFrags"], 1),
    ([-1, "spliceType"], "ONLY_REF_SPLICE"),
    ([-1, "left", "geneName"], "NCOR2"),
    ([-1, "left", "geneID"], "ENSG00000196498.13"),
    ([-1, "left", "chromosome"], "chr12"),
    ([-1, "left", "position"], 124362126),
    ([-1, "left", "strand"], "-"),
    ([-1, "right", "geneName"], "UBC"),
    ([-1, "right", "geneID"], "ENSG00000150991.14"),
    ([-1, "right", "chromosome"], "chr12"),
    ([-1, "right", "position"], 124913774),
    ([-1, "right", "strand"], "-"),
    ([-1, "reads", "junctionReads", 0],
     "HISEQ:113:C6ALHANXX:5:1310:3755:40412"),
    ([-1, "reads", "junctionReads", -1],
     "HISEQ:113:C6ALHANXX:5:1310:3755:40412"),
    ([-1, "reads", "spanningFrags", 0],
     "HISEQ:114:C6DWCANXX:4:2101:4289:41779"),
    ([-1, "reads", "spanningFrags", -1],
     "HISEQ:114:C6DWCANXX:4:2101:4289:41779"),
])
def test_star_fusion_v060_02(star_fusion_v060_02, attrs, exp):
    assert getattr_nested(star_fusion_v060_02.json, attrs) == exp, \
        ", ".join([repr(x) for x in attrs])



# -*- coding: utf-8 -*-
"""
    crimson.tests.utils
    ~~~~~~~~~~~~~~~~~~~

    General test utilities.

"""
# (c) 2015-2018 Wibowo Arindrarto <bow@bow.web.id>
from os.path import abspath, dirname, join


TEST_CASE_DIR = abspath(join(dirname(__file__), "cases"))


def get_test_path(bname, test_dir=TEST_CASE_DIR):
    """Helper method to return the path of a test case file or directory.

    :param bname: Test case base name.
    :type bname: str
    :param test_dir: Test case directory name.
    :type test_dir: str
    :returns: Absolute path to the test case file or directory.
    :rtype: str

    """
    return abspath(join(test_dir, bname))


def getattr_nested(obj, idxs):
    """Helper method for recursively fetching an item from the given object.

    :param obj: Object containing the item to retrieve.
    :type obj: object
    :param idxs: List of attribute names / indices / keys to retrieve the item.
    :type idxs: list(str, int)
    :returns: None or the item.

    """
    if len(idxs) == 0:
        return obj

    idx = idxs.pop(0)

    if isinstance(obj, dict):
        if idx in obj:
            return getattr_nested(obj[idx], idxs)
    elif isinstance(obj, (list, tuple)) and isinstance(idx, int):
        if idx < len(obj):
            return getattr_nested(obj[idx], idxs)
    else:
        return getattr_nested(getattr(obj, idx))



# -*- coding: utf-8 -*-
"""
    fusioncatcher subcommand tests
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
# (c) 2015-2018 Wibowo Arindrarto <bow@bow.web.id>
import json
import pytest
from click.testing import CliRunner

from crimson.cli import main

from .utils import get_test_path, getattr_nested


@pytest.fixture(scope="module")
def fusioncatcher_fail():
    runner = CliRunner()
    in_file = get_test_path("fusioncatcher_nope.txt")
    result = runner.invoke(main, ["fusioncatcher", in_file])
    return result


@pytest.fixture(scope="module")
def fusioncatcher_v0995a():
    runner = CliRunner()
    in_file = get_test_path("fusioncatcher_v0995a.txt")
    result = runner.invoke(main, ["fusioncatcher", in_file])
    result.json = json.loads(result.output)
    return result


def test_fusioncatcher_fail_exit_code(fusioncatcher_fail):
    assert fusioncatcher_fail.exit_code != 0


def test_fusioncatcher_fail_output(fusioncatcher_fail):
    err_msg = "Unexpected column names:"
    assert err_msg in fusioncatcher_fail.output


@pytest.mark.parametrize("attrs, exp", [
    ([0, "5end", "geneSymbol"], "RUNX1"),
    ([0, "5end", "geneID"], "ENSG00000159216"),
    ([0, "5end", "exonID"], "ENSE00003512550"),
    ([0, "5end", "chromosome"], "21"),
    ([0, "5end", "position"], 34859474),
    ([0, "5end", "strand"], "-"),
    ([0, "3end", "geneSymbol"], "RUNX1T1"),
    ([0, "3end", "geneID"], "ENSG00000079102"),
    ([0, "3end", "exonID"], "ENSE00003614817"),
    ([0, "3end", "chromosome"], "8"),
    ([0, "3end", "position"], 92017363),
    ([0, "3end", "strand"], "-"),
    ([0, "fusionDescription"], ["known", "chimerdb2", "ticdb", "tcga",
                                "cell_lines"]),
    ([0, "nCommonMappingReads"], 0),
    ([0, "nSpanningPairs"], 41),
    ([0, "nSpanningUniqueReads"], 16),
    ([0, "longestAnchorLength"], 62),
    ([0, "fusionFindingMethod"], ["BOWTIE", "BOWTIE+BLAT", "BOWTIE+STAR"]),
    ([0, "fusionSequence"],
        "CTACCACAGAGCCATCAAAATCACAGTGGATGGGCCCCGAGAACCTCGAA*ATCGTACTGAGAAGCA"
        "CTCCACAATGCCAGACTCACCTGTGGATGTGAAG"),
    ([0, "predictedEffect"], "in-frame"),
    ([0, "predictedFusedTranscripts", 0],
        "ENST00000437180:803/ENST00000522467:311"),
    ([0, "predictedFusedTranscripts", -1],
        "ENST00000344691:2110/ENST00000518992:377"),
    ([0, "predictedFusedProteins", 0],
        "MASDSIFESFPSYPQCFMRECILGMNPSRDVHDASTSRRFTPPSTALSPGKMSEALPLGAPDAGAAL"
        "AGKLRSGDRSMVEVLADHPGELVRTDSPNFLCSVLPTHWRCNKTLPIAFKVVALGDVPDGTLVTVMA"
        "GNDENYSAELRNATAAMKNQVARFNDLRFVGRSGRGKSFTLTITVFTNPPQVATYHRAIKITVDGPR"
        "EPRNRTEKHSTMPDSPVDVKTQSRLTPPTMPPPPTTQGAPRTSSFTPTTLTNGTSHSPTALNGAPSP"
        "PNGFSNGPSSSSSSSLANQQLPPACGARQLSKLKRFLTTLQQFGNDISPEIGERVRTLVLGLV"),
    ([0, "predictedFusedProteins", -1],
        "MRIPVDASTSRRFTPPSTALSPGKMSEALPLGAPDAGAALAGKLRSGDRSMVEVLADHPGELVRTDS"
        "PNFLCSVLPTHWRCNKTLPIAFKVVALGDVPDGTLVTVMAGNDENYSAELRNATAAMKNQVARFNDL"
        "RFVGRSGRGKSFTLTITVFTNPPQVATYHRAIKITVDGPREPRNRTEKHSTMPDSPVDVKTQSRLTP"
        "PTMPPPPTTQGAPRTSSFTPTTLTNGTSHSPTALNGAPSPPNGFSNGPSSSSSSSLANQQLPPACGA"
        "RQLSKLKRFLTTLQQFGNDISPEIGERVRTLVLGLVNSTLTIEEFHSKLQEATNFPLRPFVIPFLKA"
        "NLPLLQRELLHCARLAKQNPAQYLAQHEQLLLDASTTSPVDS"),

    ([-1, "5end", "geneSymbol"], "VPS45"),
    ([-1, "5end", "geneID"], "ENSG00000136631"),
    ([-1, "5end", "exonID"], "ENSE00003679462"),
    ([-1, "5end", "chromosome"], "1"),
    ([-1, "5end", "position"], 150110627),
    ([-1, "5end", "strand"], "+"),
    ([-1, "3end", "geneSymbol"], "PLEKHO1"),
    ([-1, "3end", "geneID"], "ENSG00000023902"),
    ([-1, "3end", "exonID"], "ENSE00003616995"),
    ([-1, "3end", "chromosome"], "1"),
    ([-1, "3end", "position"], 150150912),
    ([-1, "3end", "strand"], "+"),
    ([-1, "fusionDescription"], ["adjacent", "known", "healthy", "hpa",
                                 "banned", "1K<gap<10K", "readthrough"]),
    ([-1, "nCommonMappingReads"], 0),
    ([-1, "nSpanningPairs"], 1),
    ([-1, "nSpanningUniqueReads"], 3),
    ([-1, "longestAnchorLength"], 29),
    ([-1, "fusionFindingMethod"], ["BOWTIE"]),
    ([-1, "fusionSequence"],
        "TGAGGATTGTCCTGGGAGGCACCACAGTGCACAACACGAAAAG*"
        "GGACCTCAGGATGGAAACCAGCAGCCTGCACCGCCCGAGAAGG"),
    ([-1, "predictedEffect"], "out-of-frame"),
    ([-1, "predictedFusedTranscripts", 0],
        "ENST00000369128:1550/ENST00000369124:309"),
    ([-1, "predictedFusedTranscripts", -1],
        "ENST00000369130:2171/ENST00000369124:309"),
    ([-1, "predictedFusedProteins", 0],
        "MVYTQSEILQKEVYLFERIDSQNREIMKHLKAICFLRPTKENVDYIIQELRRPKYTIYFIYFSNVISK"
        "SDVKSLAEADEQEVVAEVQQVITKEYELFEFRRTEVPPLLLILDRCDDAITPLLNQWTYQAMVHELLG"
        "INNNRIDLSRVPGISKDLREVVLSAENDEFYANNMYLNFAEIGSNIKNLMEDFQKKKPKEQQKLESIA"
        "DMKAFVENYPQFKKMSGTVSKHVTVVGELSRLVSERNLLEVSEVEQELACQNDHSSALQNIKRLLQNP"
        "KVTEFDAARLVMLYALHYERHSSNSLPGLMMDLRNKGVSEKYRKLVSAVVEYGGKRVRGSDLFSPKDA"
        "VAITKQFLKGLKGVENVYTQHQPFLHETLDHLIKGRLKENLYPYLGPSTLRDRPQDIIVFVIGGATYE"
        "EALTVYNLNRTTPGVRIVLGGTTVHNTKRDLRMETSSLHRPRRSAGSGNSAGKGFSGRFGKTAMWC"),
    ([-1, "predictedFusedProteins", -1],
        "MNVVFAVKQYISKMIEDSGPGMKVLLMDKETTGIVSMVYTQSEILQKEVYLFERIDSQNREIMKHLKA"
        "ICFLRPTKENVDYIIQELRRPKYTIYFIYFSNVISKSDVKSLAEADEQEVVAEVQEFYGDYIAVNPHL"
        "FSLNILGCCQGRNWDPAQLSRTTQGLTALLLSLKKCPMIRYQLSSEAAKRLAECVKQVITKEYELFEF"
        "RRTEVPPLLLILDRCDDAITPLLNQWTYQAMVHELLGINNNRIDLSRVPGISKDLREVVLSAENDEFY"
        "ANNMYLNFAEIGSNIKNLMEDFQKKKPKEQQKLESIADMKAFVENYPQFKKMSGTVSKHVTVVGELSR"
        "LVSERNLLEVSEVEQELACQNDHSSALQNIKRLLQNPKVTEFDAARLVMLYALHYERHSSNSLPGLMM"
        "DLRNKGVSEKYRKLVSAVVEYGGKRVRGSDLFSPKDAVAITKQFLKGLKGVENVYTQHQPFLHETLDH"
        "LIKGRLKENLYPYLGPSTLRDRPQDIIVFVIGGATYEEALTVYNLNRTTPGVRIVLGGTTVHNTKRDL"
        "RMETSSLHRPRRSAGSGNSAGKGFSGRFGKTAMWC"),
])
def test_fusioncatcher_v0995a(fusioncatcher_v0995a, attrs, exp):
    assert getattr_nested(fusioncatcher_v0995a.json, attrs) == exp, \
        ", ".join([repr(x) for x in attrs])



# -*- coding: utf-8 -*-
"""
    star subcommand tests
    ~~~~~~~~~~~~~~~~~~~~~

"""
# (c) 2015-2018 Wibowo Arindrarto <bow@bow.web.id>
import json
import pytest
from click.testing import CliRunner

from crimson.cli import main

from .utils import get_test_path


@pytest.fixture(scope="module")
def star_fail():
    runner = CliRunner()
    in_file = get_test_path("star_nope.txt")
    result = runner.invoke(main, ["star", in_file])
    return result


@pytest.fixture(scope="module")
def star_v230_01():
    runner = CliRunner()
    in_file = get_test_path("star_v230_01.txt")
    result = runner.invoke(main, ["star", in_file])
    result.json = json.loads(result.output)
    return result


@pytest.fixture(scope="module")
def star_v230_02():
    runner = CliRunner()
    in_file = get_test_path("star_v230_02.txt")
    result = runner.invoke(main, ["star", in_file])
    result.json = json.loads(result.output)
    return result


def test_star_fail_exit_code(star_fail):
    assert star_fail.exit_code != 0


def test_star_fail_output(star_fail):
    err_msg = "Unexpected file structure. No contents parsed."
    assert err_msg in star_fail.output


@pytest.mark.parametrize("attr, exp", [
    ("avgDeletionLength", 1.36),
    ("avgInputLength", 98),
    ("avgInsertionLength", 1.21),
    ("avgMappedLength", 98.27),
    ("mappingSpeed", 403.16),
    ("nInput", 14782416),
    ("nMappedMultipleLoci", 1936775),
    ("nMappedTooManyLoci", 27644),
    ("nSplicesATAC", 2471),
    ("nSplicesAnnotated", 3780876),
    ("nSplicesGCAG", 22344),
    ("nSplicesGTAG", 3780050),
    ("nSplicesNonCanonical", 5148),
    ("nSplicesTotal", 3810013),
    ("nUniquelyMapped", 12347431),
    ("pctMappedMultipleLoci", 13.1),
    ("pctMappedTooManyLoci", 0.19),
    ("pctUniquelyMapped", 83.53),
    ("pctUnmappedForOther", 0.03),
    ("pctUnmappedForTooManyMismatches", 0.0),
    ("pctUnmappedForTooShort", 3.16),
    ("rateDeletionPerBase", 0.0),
    ("rateInsertionPerBase", 0.0),
    ("rateMismatchPerBase", 0.24),
    ("timeEnd", "Dec 11 19:01:56"),
    ("timeJobStart", "Dec 11 18:55:02"),
    ("timeMappingStart", "Dec 11 18:59:44"),
])
def test_star_v230_01(star_v230_01, attr, exp):
    assert star_v230_01.json.get(attr) == exp, attr



# -*- coding: utf-8 -*-
"""
    fastqc subcommand tests
    ~~~~~~~~~~~~~~~~~~~~~~~

"""
# (c) 2015-2018 Wibowo Arindrarto <bow@bow.web.id>
import json
import os

import pytest
from click.testing import CliRunner

from crimson.cli import main
from .utils import get_test_path, getattr_nested


def test_fastqc_dir_error():
    runner = CliRunner()
    result = runner.invoke(main, ["fastqc", os.getcwd()])
    assert result.exit_code != 0


@pytest.fixture(scope="module")
def fastqc_fail():
    runner = CliRunner()
    in_file = get_test_path("fastqc_nope.txt")
    result = runner.invoke(main, ["fastqc", in_file])
    return result


def test_fastqc_fail(fastqc_fail):
    assert fastqc_fail.exit_code == -1


@pytest.fixture(scope="module")
def fastqc_v0101_01():
    runner = CliRunner()
    in_file = get_test_path("fastqc_v0101_01.txt")
    result = runner.invoke(main, ["fastqc", in_file])
    result.json = json.loads(result.output)
    return result


def test_fastqc_v0101_01_exit_code(fastqc_v0101_01):
    assert fastqc_v0101_01.exit_code == 0


@pytest.mark.parametrize("attrs, exp", [
    (["version"], "0.10.1"),
])
def test_fastqc_v0101_01_root(fastqc_v0101_01, attrs, exp):
    # since the function call modifies the list
    ori_attrs = list(attrs)
    assert getattr_nested(fastqc_v0101_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in ori_attrs])


@pytest.mark.parametrize("attrs, exp", [
    (["status"], "pass"),
    (["contents", "Filename"], "/home/crimson/fastqc/input.fq.gz"),
    (["contents", "File type"], "Conventional base calls"),
    (["contents", "Encoding"], "Sanger / Illumina 1.9"),
    (["contents", "Total Sequences"], 7154316),
    (["contents", "Filtered Sequences"], 0),
    (["contents", "Sequence length"], 125),
    (["contents", "%GC"], 48),
])
def test_fastqc_v0101_01_basic_statistics(fastqc_v0101_01, attrs, exp):
    attrs = ["Basic Statistics"] + attrs
    ori_attrs = list(attrs)
    assert getattr_nested(fastqc_v0101_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in ori_attrs])


@pytest.mark.parametrize("attrs, exp", [
    (["status"], "pass"),
    (["contents", 0, "Base"], "1"),
    (["contents", 0, "Mean"], 26.173029818643737),
    (["contents", 0, "Median"], 28.0),
    (["contents", 0, "Lower Quartile"], 18.0),
    (["contents", 0, "Upper Quartile"], 32.0),
    (["contents", 0, "10th Percentile"], 18.0),
    (["contents", 0, "90th Percentile"], 33.0),
    (["contents", 124, "Base"], "125"),
    (["contents", 124, "Mean"], 33.835322482261056),
    (["contents", 124, "Median"], 38.0),
    (["contents", 124, "Lower Quartile"], 35.0),
    (["contents", 124, "Upper Quartile"], 38.0),
    (["contents", 124, "10th Percentile"], 25.0),
    (["contents", 124, "90th Percentile"], 38.0),
    (["contents", 125], None),
])
def test_fastqc_v0101_01_base_sequence_quality(fastqc_v0101_01, attrs, exp):
    attrs = ["Per base sequence quality"] + attrs
    ori_attrs = list(attrs)
    assert getattr_nested(fastqc_v0101_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in ori_attrs])


@pytest.mark.parametrize("attrs, exp", [
    (["status"], "pass"),
    (["contents", 0, "Quality"], 2),
    (["contents", 0, "Count"], 522.0),
    (["contents", 35, "Quality"], 37),
    (["contents", 35, "Count"], 4126023.0),
    (["contents", 36], None),
])
def test_fastqc_v0101_01_sequence_quality_scores(fastqc_v0101_01, attrs, exp):
    attrs = ["Per sequence quality scores"] + attrs
    ori_attrs = list(attrs)
    assert getattr_nested(fastqc_v0101_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in ori_attrs])


@pytest.mark.parametrize("attrs, exp", [
    (["status"], "fail"),
    (["contents", 0, "Base"], "1"),
    (["contents", 0, "G"], 45.23537595867585),
    (["contents", 0, "A"], 16.85510270835236),
    (["contents", 0, "T"], 12.633971292200583),
    (["contents", 0, "C"], 25.27555004077121),
    (["contents", 124, "Base"], "125"),
    (["contents", 124, "G"], 24.28912421870512),
    (["contents", 124, "A"], 25.548183619130793),
    (["contents", 124, "T"], 25.64126039286254),
    (["contents", 124, "C"], 24.52143176930155),
    (["contents", 125], None),
])
def test_fastqc_v0101_01_base_sequence_content(fastqc_v0101_01, attrs, exp):
    attrs = ["Per base sequence content"] + attrs
    ori_attrs = list(attrs)
    assert getattr_nested(fastqc_v0101_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in ori_attrs])


@pytest.mark.parametrize("attrs, exp", [
    (["status"], "fail"),
    (["contents", 0, "Base"], "1"),
    (["contents", 0, "%GC"], 70.51092599944707),
    (["contents", 124, "Base"], "125"),
    (["contents", 124, "%GC"], 48.81055598800667),
    (["contents", 125], None),
])
def test_fastqc_v0101_01_base_gc(fastqc_v0101_01, attrs, exp):
    attrs = ["Per base GC content"] + attrs
    ori_attrs = list(attrs)
    assert getattr_nested(fastqc_v0101_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in ori_attrs])


@pytest.mark.parametrize("attrs, exp", [
    (["status"], "warn"),
    (["contents", 0, "GC Content"], 0),
    (["contents", 0, "Count"], 2.0),
    (["contents", 100, "GC Content"], 100),
    (["contents", 100, "Count"], 0.0),
    (["contents", 101], None),
])
def test_fastqc_v0101_01_sequence_gc(fastqc_v0101_01, attrs, exp):
    attrs = ["Per sequence GC content"] + attrs
    ori_attrs = list(attrs)
    assert getattr_nested(fastqc_v0101_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in ori_attrs])


@pytest.mark.parametrize("attrs, exp", [
    (["status"], "pass"),
    (["contents", 0, "Base"], "1"),
    (["contents", 0, "N-Count"], 0.0480409308171459),
    (["contents", 124, "Base"], "125"),
    (["contents", 124, "N-Count"], 9.784303628746619e-5),
    (["contents", 125], None),
])
def test_fastqc_v0101_01_base_n_content(fastqc_v0101_01, attrs, exp):
    attrs = ["Per base N content"] + attrs
    ori_attrs = list(attrs)
    assert getattr_nested(fastqc_v0101_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in ori_attrs])


@pytest.mark.parametrize("attrs, exp", [
    (["status"], "pass"),
    (["contents", 0, "Length"], 125),
    (["contents", 0, "Count"], 7154316.0),
    (["contents", 1], None),
])
def test_fastqc_v0101_01_length_distribution(fastqc_v0101_01, attrs, exp):
    attrs = ["Sequence Length Distribution"] + attrs
    ori_attrs = list(attrs)
    assert getattr_nested(fastqc_v0101_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in ori_attrs])


@pytest.mark.parametrize("attrs, exp", [
    (["status"], "fail"),
    (["Total Duplicate Percentage"], 66.47973811024582),
    (["contents", 0, "Duplication Level"], 1),
    (["contents", 0, "Relative count"], 100.0),
    (["contents", 9, "Duplication Level"], "10++"),
    (["contents", 9, "Relative count"], 37.5636919493376),
    (["contents", 10], None),
])
def test_fastqc_v0101_01_duplication_levels(fastqc_v0101_01, attrs, exp):
    attrs = ["Sequence Duplication Levels"] + attrs
    ori_attrs = list(attrs)
    assert getattr_nested(fastqc_v0101_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in ori_attrs])


@pytest.mark.parametrize("attrs, exp", [
    (["status"], "pass"),
    (["contents", 0], None),
])
def test_fastqc_v0101_01_overrepresented(fastqc_v0101_01, attrs, exp):
    attrs = ["Overrepresented sequences"] + attrs
    ori_attrs = list(attrs)
    assert getattr_nested(fastqc_v0101_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in ori_attrs])


@pytest.mark.parametrize("attrs, exp", [
    (["status"], "warn"),
    (["contents", 0, "Sequence"], "TTTTT"),
    (["contents", 0, "Count"], 3280715),
    (["contents", 0, "Obs/Exp Overall"], 3.415043),
    (["contents", 0, "Obs/Exp Max"], 5.682842),
    (["contents", 0, "Max Obs/Exp Position"], 25),
    (["contents", 93, "Sequence"], "GTACG"),
    (["contents", 93, "Count"], 370820),
    (["contents", 93, "Obs/Exp Overall"], 0.4485244),
    (["contents", 93, "Obs/Exp Max"], 6.9574623),
    (["contents", 93, "Max Obs/Exp Position"], 14),
    (["contents", 94], None),
])
def test_fastqc_v0101_01_kmer_content(fastqc_v0101_01, attrs, exp):
    attrs = ["Kmer Content"] + attrs
    ori_attrs = list(attrs)
    assert getattr_nested(fastqc_v0101_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in ori_attrs])


def test_fastqc_v0101_01_dir():
    runner = CliRunner()
    dir_name = "fastqc_v0101_01_dir.fastqc"
    dir_path = get_test_path(dir_name)
    dir_result = runner.invoke(main, ["fastqc", dir_path])

    file_name = "fastqc_v0101_01_dir.fastqc/input.fq_fastqc/fastqc_data.txt"
    file_path = get_test_path(file_name)
    file_result = runner.invoke(main, ["fastqc", file_path])

    assert dir_result.output == file_result.output


def test_fastqc_v0101_02_zip():
    runner = CliRunner()
    zip_path = get_test_path("fastqc_v0101_02.fq_fastqc.zip")
    zip_result = runner.invoke(main, ["fastqc", zip_path])

    data_path = get_test_path("fastqc_v0101_02.txt")
    data_result = runner.invoke(main, ["fastqc", data_path])

    assert data_result.output
    assert data_result.output == zip_result.output


@pytest.fixture(scope="module")
def fastqc_zip_not_fastqc():
    runner = CliRunner()
    in_file = get_test_path("fastqc_not_fastqc.zip")
    result = runner.invoke(main, ["fastqc", in_file])
    return result


def test_fastqc_zip_not_fastqc(fastqc_zip_not_fastqc):
    assert fastqc_zip_not_fastqc.exit_code != 0
    print(dir(fastqc_zip_not_fastqc))
    assert "contains an unexpected number of" in fastqc_zip_not_fastqc.output



# -*- coding: utf-8 -*-
"""
    crimson
    ~~~~~~~

    :license: BSD

    Converter for various bioinformatics tool outputs.

"""
# (c) 2015-2018 Wibowo Arindrarto <bow@bow.web.id>

RELEASE = True

__version_info__ = ("0", "4", "0")
__version__ = ".".join(__version_info__)
__version__ += "-dev" if not RELEASE else ""

__author__ = "Wibowo Arindrarto"
__contact__ = "bow@bow.web.id"
__homepage__ = "http://bow.web.id"



# -*- coding: utf-8 -*-
"""
    crimson.fastqc
    ~~~~~~~~~~~~~~

    FastQC output parsing.

"""
# (c) 2015-2018 Wibowo Arindrarto <bow@bow.web.id>
from io import StringIO
from os import path, walk
from zipfile import is_zipfile, ZipFile

import click

from .utils import convert, get_handle


__all__ = ["parse"]

_MAX_SIZE = 1024 * 1024 * 10
_MAX_LINE_SIZE = 1024
_RESULTS_FNAME = "fastqc_data.txt"


class FastQCModule(object):

    """Class representing a FastQC analysis module."""

    def __init__(self, raw_lines, end_mark='>>END_MODULE'):
        """

        :param raw_lines: list of lines in the module
        :type raw_lines: list of str
        :param end_mark: mark of the end of the module
        :type end_mark: str

        """
        self.extra = {}
        self.raw_lines = raw_lines
        self.end_mark = end_mark
        self.status = None
        self.name = None
        self.contents = self._parse()

    @property
    def dict(self):
        """Module data as a dictionary."""
        payload = {
            "contents": self.contents,
            "status": self.status
        }
        if len(self.extra) > 0:
            for ek, ev in self.extra.items():
                payload[ek] = ev
        return payload

    def _parse(self):
        """Common parser for a FastQC module."""

        # Helper function for converting FastQC values that keeps
        # the "Base" column as strings (since it can be a number
        # or a strin denoting a range)
        def fqc_convert(k, v):
            if k == "Base":
                return v
            return convert(v)

        # check that the last line is a proper end mark
        assert self.raw_lines[-1].startswith(self.end_mark)
        # parse name and status from first line
        tokens = self.raw_lines[0].strip().split('\t')
        name = tokens[0][2:]
        self.name = name
        status = tokens[-1]
        self.status = status
        # the rest of the lines except the last one
        data = []
        if self.name != "Sequence Duplication Levels":
            # and column names from second/third line
            columns = self.raw_lines[1][1:].strip().split("\t")
            self._columns = columns
            for line in self.raw_lines[2:-1]:
                cols = line.strip().split("\t")
                data.append(cols)
        else:
            extra_k, extra_v = self.raw_lines[1][1:].strip().split("\t")
            self.extra[extra_k] = convert(extra_v)
            columns = self.raw_lines[2][1:].strip().split("\t")
            self._columns = columns
            for line in self.raw_lines[3:-1]:
                cols = line.strip().split("\t")
                data.append(cols)

        # optional processing for different modules
        if self.name == 'Basic Statistics':
            data = {k: convert(v) for k, v in data}
        else:
            # zip column names and its values ~ each item in array == one row
            data = [zip(columns, [v for v in d]) for d in data]
            # try to convert numbers appropriately
            # except for "Base" column, since FastQC may output it as range
            data = [{k: fqc_convert(k, v) for k, v in zpd} for zpd in data]

        return data


class FastQC(object):

    """Class representing results from a FastQC run."""

    _mod_names = [
        ">>Basic Statistics",
        ">>Per base sequence quality",
        ">>Per sequence quality scores",
        ">>Per base sequence content",
        ">>Per base GC content",
        ">>Per sequence GC content",
        ">>Per base N content",
        ">>Sequence Length Distribution",
        ">>Sequence Duplication Levels",
        ">>Overrepresented sequences",
        ">>Kmer Content",
    ]

    _mod_map = {k: k.lstrip(">") for k in _mod_names}

    def __init__(self, fp, max_size=_MAX_SIZE, max_line_size=_MAX_LINE_SIZE):
        """

        :param fp: open file handle pointing to the FastQC data file
        :type fp: file handle
        :param max_size: Maximum allowed size of the FastQC data file
                        (default: 10 MiB).
        :type max_size: int
        :param max_line_size: maximum number of bytes read everytime the
                              underlying ``readline`` is called (default: 1024).
        :type max_line_size: int

        """
        self.modules = {}
        self._max_line_size = _MAX_LINE_SIZE

        line = fp.readline(self._max_line_size)
        attr = ""
        read_size = self._max_line_size
        while read_size <= max_size:

            tokens = line.strip().split('\t')
            # break on EOF
            if not line:
                break
            # parse version
            elif line.startswith('##FastQC'):
                self.version = line.strip().split()[1]
            # parse individual modules
            elif tokens[0] in self._mod_map:
                attr = self._mod_map[tokens[0]]
                raw_lines = self._read_module(fp, line)
                self.modules[attr] = FastQCModule(raw_lines)

            line = fp.readline(self._max_line_size)
            read_size += self._max_line_size

    def _read_module(self, fp, line):
        """Returns a list of lines in a module.

        :param fp: open file handle pointing to the FastQC data file
        :type fp: file handle
        :param line: first line in the module
        :type line: str
        :returns: a list of lines in the module
        :rtype: list of str

        """
        raw = [line]
        while not line.startswith('>>END_MODULE'):
            line = fp.readline(self._max_line_size)
            raw.append(line)

            if not line:
                raise ValueError("Unexpected end of file in module %r" % line)

        return raw

    @property
    def dict(self):
        """FastQC data as a dictionary."""
        payload = {k: v.dict for k, v in self.modules.items()}
        payload["version"] = self.version
        return payload


def parse(in_data, encoding="utf-8", results_fname=_RESULTS_FNAME,
          max_size=_MAX_SIZE):
    """Parses FastQC results into a dictionary.

    :param in_data: File handle of a fastqc_data.txt file, or path to a
                    fastqc_data.txt file, or path to a FastQC results
                    directory, or path to a zipped FastQC result.
    :type in_data: str or file handle
    :param encoding: Encoding of the input file. This is ignored if ``in_data``
                     is a file handle (default: utf-8).
    :type encoding: string
    :param results_fname: Name of the text file produced by FastQC in which all
                          the results are stored. This is ignored if ``in_data``
                          is a file handle (default: fastqc_data.txt).
    :param max_size: Maximum allowed size of the FastQC data file
                     (default: 10 MiB).
    :type max_size: int
    :returns: Parsed FastQC values.
    :rtype: dict

    """
    # Input is zipped FastQC result, extract data file contents into a file-like
    # handle and parse it.
    if is_zipfile(in_data):
        zf = ZipFile(in_data)
        try:
            data_fname, = [f for f in zf.namelist()
                           if f.endswith(results_fname)]
        except ValueError:
            raise click.BadParameter("File {0} contains an unexpected number"
                                     " of files named {1}."
                                     "".format(in_data, results_fname))
        with zf.open(data_fname) as src:
            data_contents = src.read(max_size).decode(encoding)
        return FastQC(StringIO(data_contents)).dict

    # Input is FastQC directory.
    if path.isdir(in_data):
        try:
            ori = in_data
            in_data = path.join(ori, next(walk(ori))[1][0], results_fname)
        except IndexError:
            raise click.BadParameter("Cannot find {0} file in the given"
                                     " directory.".format(results_fname))

    # Input is a fastqc_data.txt file handle or path to it.
    with get_handle(in_data, encoding=encoding) as fh:
        return FastQC(fh, max_size=max_size).dict



# -*- coding: utf-8 -*-
"""
    flagstat subcommand tests
    ~~~~~~~~~~~~~~~~~~~~~~~~~

"""
# (c) 2015-2018 Wibowo Arindrarto <bow@bow.web.id>
import json
import pytest
from click.testing import CliRunner

from crimson.cli import main

from .utils import get_test_path


@pytest.fixture(scope="module")
def flagstat_fail():
    runner = CliRunner()
    in_file = get_test_path("samtools_flagstat_nope.txt")
    result = runner.invoke(main, ["flagstat", in_file])
    return result


@pytest.fixture(scope="module")
def flagstat_v0119_01():
    runner = CliRunner()
    in_file = get_test_path("samtools_flagstat_v0119_01.txt")
    result = runner.invoke(main, ["flagstat", in_file])
    result.json = json.loads(result.output)
    return result


@pytest.fixture(scope="module")
def flagstat_v11_01():
    runner = CliRunner()
    in_file = get_test_path("samtools_flagstat_v11_01.txt")
    result = runner.invoke(main, ["flagstat", in_file])
    result.json = json.loads(result.output)
    return result


def test_flagstat_fail_exit_code(flagstat_fail):
    assert flagstat_fail.exit_code != 0


def test_flagstat_fail_output(flagstat_fail):
    err_msg = "Cannot parse input flagstat file."
    assert err_msg in flagstat_fail.output


def test_flagstat_v0119_01_exit_code(flagstat_v0119_01):
    assert flagstat_v0119_01.exit_code == 0


@pytest.mark.parametrize("attr, exp", [
    ("total", 14152593),
    ("duplicates", 0),
    ("mapped", 13391622),
    ("paired_sequencing", 14152593),
    ("paired", 13356881),
    ("paired_proper", 13331891),
    ("read1", 7076309),
    ("read2", 7076284),
    ("singleton", 34741),
    ("diff_chrom", 26959),
    ("diff_chrom_mapq", 26959),
])
def test_flagstat_v0119_01_pass_qc(flagstat_v0119_01, attr, exp):
    assert flagstat_v0119_01.json.get("pass_qc", {}).get(attr) == exp


@pytest.mark.parametrize("attr, exp", [
    ("total", 0),
    ("duplicates", 0),
    ("mapped", 0),
    ("paired_sequencing", 0),
    ("paired", 0),
    ("paired_proper", 0),
    ("read1", 0),
    ("read2", 0),
    ("singleton", 0),
    ("diff_chrom", 0),
    ("diff_chrom_mapq", 0),
])
def test_flagstat_v0119_01_fail_qc(flagstat_v0119_01, attr, exp):
    assert flagstat_v0119_01.json.get("fail_qc", {}).get(attr) == exp


def test_flagstat_v11_01_exit_code(flagstat_v11_01):
    assert flagstat_v11_01.exit_code == 0


@pytest.mark.parametrize("attr, exp", [
    ("total", 71511),
    ("secondary", 122),
    ("supplementary", 0),
    ("duplicates", 0),
    ("mapped", 71228),
    ("paired_sequencing", 71389),
    ("paired", 71070),
    ("paired_proper", 70374),
    ("read1", 35691),
    ("read2", 35698),
    ("singleton", 36),
    ("diff_chrom", 446),
    ("diff_chrom_mapq", 192),
])
def test_flagstat_v11_01_pass_qc(flagstat_v11_01, attr, exp):
    assert flagstat_v11_01.json.get("pass_qc", {}).get(attr) == exp


@pytest.mark.parametrize("attr, exp", [
    ("total", 0),
    ("secondary", 0),
    ("supplementary", 0),
    ("duplicates", 0),
    ("mapped", 0),
    ("paired_sequencing", 0),
    ("paired", 0),
    ("paired_proper", 0),
    ("read1", 0),
    ("read2", 0),
    ("singleton", 0),
    ("diff_chrom", 0),
    ("diff_chrom_mapq", 0),
])
def test_flagstat_v11_01_fail_qc(flagstat_v11_01, attr, exp):
    assert flagstat_v11_01.json.get("fail_qc", {}).get(attr) == exp



# -*- coding: utf-8 -*-
"""
    vep subcommand tests
    ~~~~~~~~~~~~~~~~~~~~

"""
# (c) 2015-2018 Wibowo Arindrarto <bow@bow.web.id>
import json
import pytest
from click.testing import CliRunner

from crimson.cli import main

from .utils import getattr_nested, get_test_path


@pytest.fixture(scope="module")
def vep_fail():
    runner = CliRunner()
    in_file = get_test_path("vep_nope.txt")
    result = runner.invoke(main, ["vep", in_file])
    return result


@pytest.fixture(scope="module")
def vep_v77_01():
    runner = CliRunner()
    in_file = get_test_path("vep_v77_01.txt")
    result = runner.invoke(main, ["vep", in_file])
    result.json = json.loads(result.output)
    return result


def test_vep_fail_exit_code(vep_fail):
    assert vep_fail.exit_code != 0


def test_vep_fail_output(vep_fail):
    err_msg = "Unexpected file structure. No contents parsed."
    assert err_msg in vep_fail.output


@pytest.mark.parametrize("attrs, exp", [
    (["VEP run statistics", "Cache/Database"],
     "/home/user/.vep/homo_sapiens/77_GRCh38"),
    (["Variant classes", "insertion"], 35),
    (["Variant classes", "deletion"], 18),
    (["Variant classes", "SNV"], 448),
    (["Variants by chromosome", "1"], 42),
    (["Variants by chromosome", "21"], 45),
    (["Position in protein", "90-100%"], 7),
    (["Distribution of variants on chromosome 21", 46], 0),
])
def test_vep_v77_01(vep_v77_01, attrs, exp):
    assert getattr_nested(vep_v77_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in attrs])



# -*- coding: utf-8 -*-
"""
    picard subcommand tests
    ~~~~~~~~~~~~~~~~~~~~~~~

"""
# (c) 2015-2018 Wibowo Arindrarto <bow@bow.web.id>
import json
import pytest
from click.testing import CliRunner

from crimson.cli import main

from .utils import get_test_path, getattr_nested


@pytest.fixture(scope="module")
def picard_fail():
    runner = CliRunner()
    in_file = get_test_path("picard_nope.txt")
    result = runner.invoke(main, ["picard", in_file])
    return result


def test_picard_fail_exit_code(picard_fail):
    assert picard_fail.exit_code != 0


@pytest.fixture(scope="module")
def alignment_summary_v1124_01():
    runner = CliRunner()
    in_file = get_test_path("picard_alignment_summary_v1124_01.txt")
    result = runner.invoke(main, ["picard", in_file])
    result.json = json.loads(result.output)
    return result


def test_alignment_summary_v1124_01_exit_code(alignment_summary_v1124_01):
    assert alignment_summary_v1124_01.exit_code == 0


@pytest.mark.parametrize("attrs, exp", [
    (["header", "time"], "Started on: Sun Jul 19 15:42:28 CEST 2015"),
    (["metrics", "class"], "picard.analysis.AlignmentSummaryMetrics"),
    (["metrics", "contents", 0, "CATEGORY"], "FIRST_OF_PAIR"),
    (["metrics", "contents", 0, "TOTAL_READS"], 35691),
    (["metrics", "contents", 0, "PF_READS"], 35691),
    (["metrics", "contents", 0, "PCT_PF_READS"], 1),
    (["metrics", "contents", 0, "PF_NOISE_READS"], 0),
    (["metrics", "contents", 0, "PF_READS_ALIGNED"], 35557),
    (["metrics", "contents", 0, "PCT_PF_READS_ALIGNED"], 0.996246),
    (["metrics", "contents", 0, "PF_ALIGNED_BASES"], 3501223),
    (["metrics", "contents", 0, "PF_HQ_ALIGNED_READS"], 33864),
    (["metrics", "contents", 0, "PF_HQ_ALIGNED_BASES"], 3346832),
    (["metrics", "contents", 0, "PF_HQ_ALIGNED_Q20_BASES"], 3334799),
    (["metrics", "contents", 0, "PF_HQ_MEDIAN_MISMATCHES"], 0),
    (["metrics", "contents", 0, "PF_MISMATCH_RATE"], 0.002714),
    (["metrics", "contents", 0, "PF_HQ_ERROR_RATE"], 0.001829),
    (["metrics", "contents", 0, "PF_INDEL_RATE"], 0.000208),
    (["metrics", "contents", 0, "MEAN_READ_LENGTH"], 99.003895),
    (["metrics", "contents", 0, "READS_ALIGNED_IN_PAIRS"], 35535),
    (["metrics", "contents", 0, "PCT_READS_ALIGNED_IN_PAIRS"], 0.999381),
    (["metrics", "contents", 0, "BAD_CYCLES"], 0),
    (["metrics", "contents", 0, "STRAND_BALANCE"], 0.500464),
    (["metrics", "contents", 0, "PCT_CHIMERAS"], 0.007275),
    (["metrics", "contents", 0, "PCT_ADAPTER"], 0.000224),
    (["metrics", "contents", 0, "SAMPLE"], ""),
    (["metrics", "contents", 0, "LIBRARY"], ""),
    (["metrics", "contents", 0, "READ_GROUP"], ""),
    (["metrics", "contents", -1, "CATEGORY"], "PAIR"),
    (["metrics", "contents", -1, "TOTAL_READS"], 71389),
    (["metrics", "contents", -1, "PF_READS"], 71389),
    (["metrics", "contents", -1, "PCT_PF_READS"], 1),
    (["metrics", "contents", -1, "PF_NOISE_READS"], 0),
    (["metrics", "contents", -1, "PF_READS_ALIGNED"], 71106),
    (["metrics", "contents", -1, "PCT_PF_READS_ALIGNED"], 0.996036),
    (["metrics", "contents", -1, "PF_ALIGNED_BASES"], 6945691),
    (["metrics", "contents", -1, "PF_HQ_ALIGNED_READS"], 67701),
    (["metrics", "contents", -1, "PF_HQ_ALIGNED_BASES"], 6638298),
    (["metrics", "contents", -1, "PF_HQ_ALIGNED_Q20_BASES"], 6607469),
    (["metrics", "contents", -1, "PF_HQ_MEDIAN_MISMATCHES"], 0),
    (["metrics", "contents", -1, "PF_MISMATCH_RATE"], 0.002845),
    (["metrics", "contents", -1, "PF_HQ_ERROR_RATE"], 0.001944),
    (["metrics", "contents", -1, "PF_INDEL_RATE"], 0.000197),
    (["metrics", "contents", -1, "MEAN_READ_LENGTH"], 98.213983),
    (["metrics", "contents", -1, "READS_ALIGNED_IN_PAIRS"], 71070),
    (["metrics", "contents", -1, "PCT_READS_ALIGNED_IN_PAIRS"], 0.999494),
    (["metrics", "contents", -1, "BAD_CYCLES"], 0),
    (["metrics", "contents", -1, "STRAND_BALANCE"], 0.500127),
    (["metrics", "contents", -1, "PCT_CHIMERAS"], 0.007275),
    (["metrics", "contents", -1, "PCT_ADAPTER"], 0.000224),
    (["metrics", "contents", -1, "SAMPLE"], ""),
    (["metrics", "contents", -1, "LIBRARY"], ""),
    (["metrics", "contents", -1, "READ_GROUP"], ""),
    (["histogram"], None),
])
def test_alignment_summary_v1124_01(alignment_summary_v1124_01, attrs, exp):
    assert getattr_nested(alignment_summary_v1124_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in attrs])


@pytest.fixture(scope="module")
def insert_size_v1124_01():
    runner = CliRunner()
    in_file = get_test_path("picard_insert_size_v1124_01.txt")
    result = runner.invoke(main, ["picard", in_file])
    result.json = json.loads(result.output)
    return result


def test_insert_size_v1124_01_exit_code(insert_size_v1124_01):
    assert insert_size_v1124_01.exit_code == 0


@pytest.mark.parametrize("attrs, exp", [
    (["header", "time"], "Started on: Sun Jul 19 15:42:28 CEST 2015"),
    (["metrics", "class"], "picard.analysis.InsertSizeMetrics"),
    (["metrics", "contents", "MEDIAN_INSERT_SIZE"], 317),
    (["metrics", "contents", "MEDIAN_ABSOLUTE_DEVIATION"], 44),
    (["metrics", "contents", "MIN_INSERT_SIZE"], 13),
    (["metrics", "contents", "MAX_INSERT_SIZE"], 23587641),
    (["metrics", "contents", "MEAN_INSERT_SIZE"], 316.523985),
    (["metrics", "contents", "STANDARD_DEVIATION"], 74.454708),
    (["metrics", "contents", "READ_PAIRS"], 35243),
    (["metrics", "contents", "PAIR_ORIENTATION"], "FR"),
    (["metrics", "contents", "WIDTH_OF_10_PERCENT"], 17),
    (["metrics", "contents", "WIDTH_OF_20_PERCENT"], 33),
    (["metrics", "contents", "WIDTH_OF_30_PERCENT"], 51),
    (["metrics", "contents", "WIDTH_OF_40_PERCENT"], 69),
    (["metrics", "contents", "WIDTH_OF_50_PERCENT"], 89),
    (["metrics", "contents", "WIDTH_OF_60_PERCENT"], 113),
    (["metrics", "contents", "WIDTH_OF_70_PERCENT"], 141),
    (["metrics", "contents", "WIDTH_OF_80_PERCENT"], 179),
    (["metrics", "contents", "WIDTH_OF_90_PERCENT"], 245),
    (["metrics", "contents", "WIDTH_OF_99_PERCENT"], 449),
    (["metrics", "contents", "SAMPLE"], ""),
    (["metrics", "contents", "LIBRARY"], ""),
    (["metrics", "contents", "READ_GROUP"], ""),
    (["histogram", "contents", 0, "insert_size"], 13),
    (["histogram", "contents", 0, "All_Reads.fr_count"], 1),
    (["histogram", "contents", -1, "insert_size"], 739),
    (["histogram", "contents", -1, "All_Reads.fr_count"], 1),
])
def test_insert_size_v1124_01(insert_size_v1124_01, attrs, exp):
    assert getattr_nested(insert_size_v1124_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in attrs])


@pytest.fixture(scope="module")
def library_complexity_v1124_01():
    runner = CliRunner()
    in_file = get_test_path("picard_library_complexity_v1124_01.txt")
    result = runner.invoke(main, ["picard", in_file])
    result.json = json.loads(result.output)
    return result


def test_library_complexity_v1124_01_exit_code(library_complexity_v1124_01):
    assert library_complexity_v1124_01.exit_code == 0


@pytest.mark.parametrize("attrs, exp", [
    (["header", "time"], "Started on: Sun Jul 19 15:42:28 CEST 2015"),
    (["metrics", "class"], "picard.sam.DuplicationMetrics"),
    (["metrics", "contents", "LIBRARY"], "Unknown"),
    (["metrics", "contents", "UNPAIRED_READS_EXAMINED"], 0),
    (["metrics", "contents", "READ_PAIRS_EXAMINED"], 35637),
    (["metrics", "contents", "UNMAPPED_READS"], 0),
    (["metrics", "contents", "UNPAIRED_READ_DUPLICATES"], 0),
    (["metrics", "contents", "READ_PAIR_DUPLICATES"], 0),
    (["metrics", "contents", "READ_PAIR_OPTICAL_DUPLICATES"], 0),
    (["metrics", "contents", "PERCENT_DUPLICATION"], 0),
    (["metrics", "contents", "ESTIMATED_LIBRARY_SIZE"], ""),
    (["histogram", "contents", 0, "duplication_group_count"], 1),
    (["histogram", "contents", 0, "Unknown"], 35637),
    (["histogram", "contents", -1, "duplication_group_count"], 7),
    (["histogram", "contents", -1, "Unknown"], 1),
])
def test_library_complexity_v1124_01(library_complexity_v1124_01, attrs, exp):
    assert getattr_nested(library_complexity_v1124_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in attrs])


@pytest.fixture(scope="module")
def mark_duplicates_v1124_01():
    runner = CliRunner()
    in_file = get_test_path("picard_mark_duplicates_v1124_01.txt")
    result = runner.invoke(main, ["picard", in_file])
    result.json = json.loads(result.output)
    return result


def test_mark_duplicates_v1124_01_exit_code(mark_duplicates_v1124_01):
    assert mark_duplicates_v1124_01.exit_code == 0


@pytest.mark.parametrize("attrs, exp", [
    (["header", "time"], "Started on: Sun Jul 19 15:42:34 CEST 2015"),
    (["metrics", "class"], "picard.sam.DuplicationMetrics"),
    (["metrics", "contents", "LIBRARY"], "SASC"),
    (["metrics", "contents", "UNPAIRED_READS_EXAMINED"], 36),
    (["metrics", "contents", "READ_PAIRS_EXAMINED"], 35535),
    (["metrics", "contents", "UNMAPPED_READS"], 283),
    (["metrics", "contents", "UNPAIRED_READ_DUPLICATES"], 0),
    (["metrics", "contents", "READ_PAIR_DUPLICATES"], 0),
    (["metrics", "contents", "READ_PAIR_OPTICAL_DUPLICATES"], 0),
    (["metrics", "contents", "PERCENT_DUPLICATION"], 0),
    (["metrics", "contents", "ESTIMATED_LIBRARY_SIZE"], ""),
    (["histogram"], None),
])
def test_mark_duplicates_v1124_01(mark_duplicates_v1124_01, attrs, exp):
    assert getattr_nested(mark_duplicates_v1124_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in attrs])


@pytest.fixture(scope="module")
def rna_seq_v1124_01():
    runner = CliRunner()
    in_file = get_test_path("picard_rna_seq_v1124_01.txt")
    result = runner.invoke(main, ["picard", in_file])
    result.json = json.loads(result.output)
    return result


def test_rna_seq_v1124_01_exit_code(rna_seq_v1124_01):
    assert rna_seq_v1124_01.exit_code == 0


@pytest.mark.parametrize("attrs, exp", [
    (["header", "time"], "Started on: Sun Jul 19 15:42:28 CEST 2015"),
    (["metrics", "class"], "picard.analysis.RnaSeqMetrics"),
    (["metrics", "contents", "PF_BASES"], 32309274),
    (["metrics", "contents", "PF_ALIGNED_BASES"], 30461277),
    (["metrics", "contents", "RIBOSOMAL_BASES"], ""),
    (["metrics", "contents", "CODING_BASES"], 10026319),
    (["metrics", "contents", "UTR_BASES"], 8691712),
    (["metrics", "contents", "INTRONIC_BASES"], 3054788),
    (["metrics", "contents", "INTERGENIC_BASES"], 8688458),
    (["metrics", "contents", "IGNORED_READS"], 0),
    (["metrics", "contents", "CORRECT_STRAND_READS"], 0),
    (["metrics", "contents", "INCORRECT_STRAND_READS"], 0),
    (["metrics", "contents", "PCT_RIBOSOMAL_BASES"], ""),
    (["metrics", "contents", "PCT_CODING_BASES"], 0.32915),
    (["metrics", "contents", "PCT_UTR_BASES"], 0.285336),
    (["metrics", "contents", "PCT_INTRONIC_BASES"], 0.100284),
    (["metrics", "contents", "PCT_INTERGENIC_BASES"], 0.28523),
    (["metrics", "contents", "PCT_MRNA_BASES"], 0.614486),
    (["metrics", "contents", "PCT_USABLE_BASES"], 0.579339),
    (["metrics", "contents", "PCT_CORRECT_STRAND_READS"], 0),
    (["metrics", "contents", "MEDIAN_CV_COVERAGE"], 0.900974),
    (["metrics", "contents", "MEDIAN_5PRIME_BIAS"], 0.048295),
    (["metrics", "contents", "MEDIAN_3PRIME_BIAS"], 0.353777),
    (["metrics", "contents", "MEDIAN_5PRIME_TO_3PRIME_BIAS"], 0.415094),
    (["metrics", "contents", "SAMPLE"], ""),
    (["metrics", "contents", "LIBRARY"], ""),
    (["metrics", "contents", "READ_GROUP"], ""),
    (["histogram", "contents", 0, "normalized_position"], 0),
    (["histogram", "contents", 0, "All_Reads.normalized_coverage"], 0.039599),
    (["histogram", "contents", -1, "normalized_position"], 100),
    (["histogram", "contents", -1, "All_Reads.normalized_coverage"], 0.142949),
])
def test_rna_seq_v1124_01(rna_seq_v1124_01, attrs, exp):
    assert getattr_nested(rna_seq_v1124_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in attrs])


@pytest.fixture(scope="module")
def wgs_v1124_01():
    runner = CliRunner()
    in_file = get_test_path("picard_wgs_v1124_01.txt")
    result = runner.invoke(main, ["picard", in_file])
    result.json = json.loads(result.output)
    return result


def test_wgs_v1124_01_exit_code(wgs_v1124_01):
    assert wgs_v1124_01.exit_code == 0


@pytest.mark.parametrize("attrs, exp", [
    (["header", "time"], "Started on: Sun Jul 19 15:42:28 CEST 2015"),
    (["metrics", "class"], "picard.analysis.CollectWgsMetrics$WgsMetrics"),
    (["metrics", "contents", "GENOME_TERRITORY"], 2867454345),
    (["metrics", "contents", "MEAN_COVERAGE"], 0.002278),
    (["metrics", "contents", "SD_COVERAGE"], 0.047962),
    (["metrics", "contents", "MEDIAN_COVERAGE"], 0),
    (["metrics", "contents", "MAD_COVERAGE"], 0),
    (["metrics", "contents", "PCT_EXC_MAPQ"], 0.044831),
    (["metrics", "contents", "PCT_EXC_DUPE"], 0),
    (["metrics", "contents", "PCT_EXC_UNPAIRED"], 0.000299),
    (["metrics", "contents", "PCT_EXC_BASEQ"], 0.004426),
    (["metrics", "contents", "PCT_EXC_OVERLAP"], 0.010468),
    (["metrics", "contents", "PCT_EXC_CAPPED"], 0),
    (["metrics", "contents", "PCT_EXC_TOTAL"], 0.060025),
    (["metrics", "contents", "PCT_5X"], 0),
    (["metrics", "contents", "PCT_10X"], 0),
    (["metrics", "contents", "PCT_15X"], 0),
    (["metrics", "contents", "PCT_20X"], 0),
    (["metrics", "contents", "PCT_25X"], 0),
    (["metrics", "contents", "PCT_30X"], 0),
    (["metrics", "contents", "PCT_40X"], 0),
    (["metrics", "contents", "PCT_50X"], 0),
    (["metrics", "contents", "PCT_60X"], 0),
    (["metrics", "contents", "PCT_70X"], 0),
    (["metrics", "contents", "PCT_80X"], 0),
    (["metrics", "contents", "PCT_90X"], 0),
    (["metrics", "contents", "PCT_100X"], 0),
    (["histogram", "contents", 0, "coverage"], 0),
    (["histogram", "contents", 0, "count"], 2860949724),
    (["histogram", "contents", -1, "coverage"], 250),
    (["histogram", "contents", -1, "count"], 0),
])
def test_wgs_v1124_01(wgs_v1124_01, attrs, exp):
    assert getattr_nested(wgs_v1124_01.json, attrs) == exp, \
        ", ".join([repr(x) for x in attrs])
